;; Generated package description from dash.el  -*- no-byte-compile: t -*-
(define-package "dash" "2.12.0.0.20171228.185742" "A modern list library for Emacs" 'nil :url "https://elpa.gnu.org/packages/dash.html" :keywords '("lists") :authors '(("Magnar Sveen" . "magnars@gmail.com")) :maintainer '("Magnar Sveen" . "magnars@gmail.com"))
